/*
 * empathy-gst-video-sink.h - Header for EmpathyGstAudioSink
 * Copyright (C) 2008 Collabora Ltd.
 * @author Sjoerd Simons <sjoerd.simons@collabora.co.uk>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef __EMPATHY_GST_AUDIO_SINK_H__
#define __EMPATHY_GST_AUDIO_SINK_H__

#include <gst/gst.h>

G_BEGIN_DECLS

typedef struct _EmpathyGstAudioSink EmpathyGstAudioSink;
typedef struct _EmpathyGstAudioSinkClass EmpathyGstAudioSinkClass;
typedef struct _EmpathyGstAudioSinkPrivate EmpathyGstAudioSinkPrivate;

struct _EmpathyGstAudioSinkClass {
    GstBinClass parent_class;
};

struct _EmpathyGstAudioSink {
    GstBin parent;
    EmpathyGstAudioSinkPrivate *priv;
};

GType empathy_audio_sink_get_type (void);

/* TYPE MACROS */
#define EMPATHY_TYPE_GST_AUDIO_SINK \
  (empathy_audio_sink_get_type ())
#define EMPATHY_GST_AUDIO_SINK(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj), EMPATHY_TYPE_GST_AUDIO_SINK, \
    EmpathyGstAudioSink))
#define EMPATHY_GST_AUDIO_SINK_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass), EMPATHY_TYPE_GST_AUDIO_SINK, \
    EmpathyGstAudioSinkClass))
#define EMPATHY_IS_GST_AUDIO_SINK(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj), EMPATHY_TYPE_GST_AUDIO_SINK))
#define EMPATHY_IS_GST_AUDIO_SINK_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass), EMPATHY_TYPE_GST_AUDIO_SINK))
#define EMPATHY_GST_AUDIO_SINK_GET_CLASS(obj) \
  (G_TYPE_INSTANCE_GET_CLASS ((obj), EMPATHY_TYPE_GST_AUDIO_SINK, \
    EmpathyGstAudioSinkClass))

GstElement *empathy_audio_sink_new (void);

void empathy_audio_sink_set_volume (EmpathyGstAudioSink *sink, gdouble volume);

gdouble empathy_audio_sink_get_volume (EmpathyGstAudioSink *sink);
void empathy_audio_sink_set_echo_cancel (EmpathyGstAudioSink *sink,
  gboolean echo_cancel);

G_END_DECLS

#endif /* #ifndef __EMPATHY_GST_AUDIO_SINK_H__*/
